@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2003, 2004, 2007, 2009, 2010, 2011
-   Free Software Foundation, Inc.
+   Copyright (C) 1999-2019 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -38,6 +37,8 @@
 #include <mailutils/header.h>
 #include <mailutils/errno.h>
 #include <mailutils/util.h>
+#include <mailutils/assoc.h>
+#include <mailutils/io.h>
 
 #include <mailutils/sys/mime.h>
 #include <mailutils/sys/message.h>
@@ -70,9 +71,9 @@ _mime_append_part (mu_mime_t mime, mu_message_t msg,
 		   size_t offset, size_t len, size_t lines)
 {
   struct _mime_part *mime_part, **part_arr;
-  int             ret;
-  size_t          size;
-  mu_header_t     hdr;
+  int ret;
+  size_t size;
+  mu_header_t hdr;
 
   if ((mime_part = calloc (1, sizeof (*mime_part))) == NULL)
     return ENOMEM;
@@ -147,8 +148,8 @@ _mime_append_part (mu_mime_t mime, mu_message_t msg,
 static void
 _mime_munge_content_header (char *field_body)
 {
-  char           *p, *e, *str = field_body;
-  int             quoted = 0;
+  char *p, *e, *str = field_body;
+  int quoted = 0;
 
   mu_str_stripws (field_body);
 
@@ -184,8 +185,8 @@ _mime_munge_content_header (char *field_body)
 static char    *
 _mime_get_param (char *field_body, const char *param, int *len)
 {
-  char           *str, *p, *v, *e;
-  int             quoted = 0, was_quoted;
+  char *str, *p, *v, *e;
+  int quoted = 0, was_quoted;
 
   if (len == NULL || (str = field_body) == NULL)
     return NULL;
@@ -242,11 +243,10 @@ _mime_append_header_line (mu_mime_t mime)
 {
   if (mime->header_length + mime->line_ndx > mime->header_buf_size)
     {
-      char           *nhb;
+      char *nhb;
 
-      if ((nhb =
-	   realloc (mime->header_buf,
-		    mime->header_length + mime->line_ndx + 128)) == NULL)
+      if ((nhb = realloc (mime->header_buf,
+			  mime->header_length + mime->line_ndx + 128)) == NULL)
 	return;
       mime->header_buf = nhb;
       mime->header_buf_size = mime->header_length + mime->line_ndx + 128;
@@ -259,20 +259,20 @@ _mime_append_header_line (mu_mime_t mime)
 static int
 _mime_parse_mpart_message (mu_mime_t mime)
 {
-  char           *cp, *cp2;
-  size_t         blength, mb_length, mb_offset, mb_lines;
-  int            ret;
-  size_t         nbytes;
+  char *cp;
+  size_t blength, mb_length, mb_offset, mb_lines;
+  int ret;
+  size_t nbytes;
 
   if (!(mime->flags & MIME_PARSER_ACTIVE))
     {
-      char           *boundary;
-      int             len;
+      char *boundary;
+      int len;
 
       if ((ret = _mime_setup_buffers (mime)) != 0)
 	return ret;
-      if ((boundary =
-	   _mime_get_param (mime->content_type, "boundary", &len)) == NULL)
+      if ((boundary = _mime_get_param (mime->content_type, "boundary", &len))
+	  == NULL)
 	return EINVAL;
       if ((mime->boundary = calloc (1, len + 1)) == NULL)
 	return ENOMEM;
@@ -289,14 +289,82 @@ _mime_parse_mpart_message (mu_mime_t mime)
   blength = strlen (mime->boundary);
 
   mu_stream_seek (mime->stream, mime->cur_offset, MU_SEEK_SET, NULL);
-  while ((ret =
-	  mu_stream_read (mime->stream, mime->cur_buf, mime->buf_size,
-			  &nbytes)) == 0 && nbytes)
+  while ((ret = mu_stream_read (mime->stream, mime->cur_buf, mime->buf_size,
+				&nbytes)) == 0 && nbytes)
     {
       cp = mime->cur_buf;
-      while (nbytes)
+
+      while (1)
 	{
-	  mime->cur_line[mime->line_ndx] = *cp;
+	  if (nbytes)
+	    mime->cur_line[mime->line_ndx] = *cp;
+
+	  if (mime->parser_state == MIME_STATE_SCAN_BOUNDARY
+	      && (nbytes == 0 || *cp == '\n'))
+	    {
+	      char *cp2 = mime->cur_line[0] == '\n'
+		            ? mime->cur_line + 1 : mime->cur_line;
+	      if (mime->line_ndx >= blength
+		  && ((memcmp (cp2, "--", 2) == 0
+		       && memcmp (cp2 + 2, mime->boundary, blength) == 0)
+		      || memcmp (cp2, mime->boundary, blength) == 0))
+		{
+		  mime->parser_state = MIME_STATE_HEADERS;
+		  mime->flags &= ~MIME_PARSER_HAVE_CR;
+		  mb_length = mime->cur_offset 
+			                   - mb_offset
+			                   - mime->line_ndx;
+		  if (mime->header_length)
+		    /* this skips the preamble */
+		    {
+		      /* RFC 1521 [Page 30]:
+			 NOTE: The CRLF preceding the encapsulation
+			 line is conceptually attached to the boundary
+			 so that it is possible to have a part that
+			 does not end with a CRLF (line break). Body
+			 parts that must be considered to end with line
+			 breaks, therefore, must have two CRLFs
+			 preceding the encapsulation line, the first
+			 of which is part of the preceding body part,
+			 and the second of which is part of the
+			 encapsulation boundary. */
+		      
+		      if (mb_lines)
+			/* to prevent negative values in case of a
+			   malformed message */
+			mb_lines--;
+		      
+		      _mime_append_part (mime, NULL,
+					 mb_offset, mb_length,
+					 mb_lines);
+		    }
+		  
+		  if ((&mime->cur_line[mime->line_ndx] - cp2 - 1 > blength
+		       && memcmp (cp2 + blength + 2, "--", 2) == 0)
+		      || (&mime->cur_line[mime->line_ndx] - cp2 - 1 == blength
+			  && memcmp (cp2 + blength, "--", 2) == 0))
+		    {	/* last boundary */
+		      mime->parser_state = MIME_STATE_BEGIN_LINE;
+		      mime->header_length = 0;
+		    }
+		  else
+		    mime->line_ndx = -1; /* headers parsing requires
+					    empty line */
+		}
+	      else if (nbytes)
+		{
+		  if (mime->header_length)
+		    mb_lines++;
+		  
+		  mime->line_ndx = 0;
+		  mime->cur_line[0] = *cp;	/* stay in this state but
+						   leave '\n' at begining */
+		}
+	    }
+
+	  if (nbytes == 0) 
+	    break;
+	  
 	  if (*cp == '\n')
 	    {
 	      switch (mime->parser_state)
@@ -308,69 +376,6 @@ _mime_parse_mpart_message (mu_mime_t mime)
 		  break;
 
 		case MIME_STATE_SCAN_BOUNDARY:
-		  cp2 =
-		    mime->cur_line[0] ==
-		    '\n' ? mime->cur_line + 1 : mime->cur_line;
-		  if (mime->line_ndx >= blength)
-		    {
-		      if ((!strncmp (cp2, "--", 2)
-			   && !mu_c_strncasecmp (cp2 + 2, mime->boundary,
-				  	         blength))
-			  || !mu_c_strncasecmp (cp2, mime->boundary, blength))
-			{
-			  mime->parser_state = MIME_STATE_HEADERS;
-			  mime->flags &= ~MIME_PARSER_HAVE_CR;
-			  mb_length = mime->cur_offset 
-			                   - mb_offset
-			                   - mime->line_ndx;
-			  if (mime->header_length)
-			    /* this skips the preamble */
-			    {
-			      /* RFC 1521 [Page 30]:
-				 NOTE: The CRLF preceding the encapsulation
-				 line is conceptually attached to the boundary
-				 so that it is possible to have a part that
-				 does not end with a CRLF (line break). Body
-				 parts that must be considered to end with line
-				 breaks, therefore, must have two CRLFs
-				 preceding the encapsulation line, the first
-				 of which is part of the preceding body part,
-				 and the second of which is part of the
-				 encapsulation boundary. */
-			      
-			      if (mb_lines)
-				/* to prevent negative values in case of a
-				   malformed message */
-				mb_lines--;
-				  
-			      _mime_append_part (mime, NULL,
-						 mb_offset, mb_length,
-						 mb_lines);
-			    }
-
-			  if ((&mime->cur_line[mime->line_ndx] - cp2 - 1 >
-			       blength
-			       && !strncmp (cp2 + blength + 2, "--", 2))
-			      || (&mime->cur_line[mime->line_ndx] - cp2 - 1 ==
-				  blength
-				  && !strncmp (cp2 + blength, "--", 2)))
-			    {	/* last boundary */
-			      mime->parser_state = MIME_STATE_BEGIN_LINE;
-			      mime->header_length = 0;
-			    }
-			  else
-			    mime->line_ndx = -1; /* headers parsing requires
-						    empty line */
-			  break;
-			}
-		    }
-
-		  if (mime->header_length)
-		    mb_lines++;
-
-		  mime->line_ndx = 0;
-		  mime->cur_line[0] = *cp;	/* stay in this state but
-						   leave '\n' at begining */
 		  break;
 
 		case MIME_STATE_HEADERS:
@@ -386,6 +391,7 @@ _mime_parse_mpart_message (mu_mime_t mime)
 		  break;
 		}
 	    }
+
 	  mime->line_ndx++;
 	  if (mime->line_ndx >= mime->line_size)
 	    {
@@ -423,7 +429,7 @@ _mime_parse_mpart_message (mu_mime_t mime)
 static int
 _mimepart_body_size (mu_body_t body, size_t *psize)
 {
-  mu_message_t       msg = mu_body_get_owner (body);
+  mu_message_t msg = mu_body_get_owner (body);
   struct _mime_part *mime_part = mu_message_get_owner (msg);
 
   if (mime_part == NULL)
@@ -436,7 +442,7 @@ _mimepart_body_size (mu_body_t body, size_t *psize)
 static int
 _mimepart_body_lines (mu_body_t body, size_t *plines)
 {
-  mu_message_t       msg = mu_body_get_owner (body);
+  mu_message_t msg = mu_body_get_owner (body);
   struct _mime_part *mime_part = mu_message_get_owner (msg);
 
   if (mime_part == NULL)
@@ -447,13 +453,20 @@ _mimepart_body_lines (mu_body_t body, size_t *plines)
 }
 
 /*------ Mime message/header functions for CREATING multipart message -----*/
+static int
+retain_charset (char const *name, void *value MU_ARG_UNUSED,
+		void *data MU_ARG_UNUSED)
+{
+  return strcmp (name, "charset") != 0;
+}
+		 
 static int
 _mime_set_content_type (mu_mime_t mime)
 {
   const char  *content_type;
-  mu_header_t     hdr = NULL;
-  size_t          size;
-  int             ret;
+  mu_header_t hdr = NULL;
+  size_t size;
+  int ret;
 
   /* Delayed the creation of the header 'til they create the final message via
      mu_mime_get_message()  */
@@ -468,7 +481,48 @@ _mime_set_content_type (mu_mime_t mime)
       if (mime->flags & MU_MIME_MULTIPART_MIXED)
 	content_type = "multipart/mixed; boundary=";
       else
-	content_type = "multipart/alternative; boundary=";
+	{
+	  size_t i;
+
+	  /* Make sure content disposition is not set for alternative
+	     parts */
+	  for (i = 0; i < mime->nmtp_parts; i++)
+	    {
+	      mu_header_t hdr;
+	      char *val;
+	      int rc;
+	      
+	      mu_message_get_header (mime->mtp_parts[i]->msg, &hdr);
+	      mu_header_remove (hdr, MU_HEADER_CONTENT_DISPOSITION, 1);
+
+	      rc = mu_header_aget_value_unfold (hdr, MU_HEADER_CONTENT_TYPE,
+						&val);
+	      if (rc == 0)
+		{
+		  mu_content_type_t ct;
+		  rc = mu_content_type_parse (val, NULL, &ct);
+		  if (rc == 0)
+		    {
+		      char *type;
+		      
+		      mu_assoc_mark (ct->param, retain_charset, NULL);
+		      mu_assoc_sweep (ct->param);
+
+		      rc = mu_asprintf (&type, "%s/%s", ct->type, ct->subtype);
+		      if (rc == 0)
+			{
+			  mu_mime_header_set (hdr,
+					      MU_HEADER_CONTENT_TYPE, type,
+					      ct->param);
+			  free (type);
+			}
+		      mu_content_type_destroy (&ct);
+		    }
+		}
+	    }
+	  
+	  content_type = "multipart/alternative; boundary=";
+	}
       if (mime->boundary == NULL)
 	{
 	  char boundary[128];
@@ -545,8 +599,11 @@ _mime_part_size (mu_mime_t mime, size_t *psize)
   int ret;
 
   if (mime->nmtp_parts == 0)
-    return EINVAL;
-
+    {
+      *psize = 0;
+      return 0;
+    }
+  
   if ((ret = _mime_set_content_type (mime)) != 0)
     return ret;
   if (mime->nmtp_parts == 1)
@@ -654,21 +711,25 @@ _mime_body_stream_seek (mu_stream_t stream, mu_off_t off, mu_off_t *presult)
   while (0)
 
 static int
-_mime_body_stream_read (mu_stream_t stream, char *buf, size_t buflen, size_t *nbytes)
+_mime_body_stream_read (mu_stream_t stream, char *buf, size_t buflen,
+			size_t *nbytes)
 {
   struct _mime_body_stream *mstr = (struct _mime_body_stream *)stream;
   mu_mime_t mime = mstr->mime;
-  int                ret = 0;
-  size_t             total = 0;
+  int ret = 0;
+  size_t total = 0;
   
   if (mime->nmtp_parts == 0)
-    return EINVAL;
+    {
+      *nbytes = 0;
+      return 0;
+    }
   
   if ((ret = _mime_set_content_type (mime)) == 0)
     {
       do
 	{
-	  size_t             part_nbytes = 0;
+	  size_t part_nbytes = 0;
 
 	  if (buflen == 0)
 	    break;
@@ -841,44 +902,63 @@ create_mime_body_stream (mu_stream_t *pstr, mu_mime_t mime)
 static int
 _mime_body_size (mu_body_t body, size_t *psize)
 {
-  mu_message_t       msg = mu_body_get_owner (body);
-  mu_mime_t          mime = mu_message_get_owner (msg);
+  mu_message_t msg = mu_body_get_owner (body);
+  mu_mime_t mime = mu_message_get_owner (msg);
   return _mime_part_size (mime, psize);
 }
 
 static int
 _mime_body_lines (mu_body_t body, size_t *plines)
 {
-  mu_message_t       msg = mu_body_get_owner (body);
-  mu_mime_t          mime = mu_message_get_owner (msg);
-  int             i, ret;
-  size_t          lines;
+  mu_message_t msg = mu_body_get_owner (body);
+  mu_mime_t mime = mu_message_get_owner (msg);
+  int i, ret;
+  size_t total = 0;
 
   if (mime->nmtp_parts == 0)
-    return EINVAL;
-
+    {
+      *plines = 0;
+      return 0;
+    }
+  
   if ((ret = _mime_set_content_type (mime)) != 0)
     return ret;
   for (i = 0; i < mime->nmtp_parts; i++)
     {
+      size_t lines;
+      
       mu_message_lines (mime->mtp_parts[i]->msg, &lines);
-      plines += lines;
+      total += lines;
       if (mime->nmtp_parts > 1)	/* boundary line */
-	plines++;
+	total++;
     }
+  *plines = total;
   return 0;
 }
 
+#define MIME_MULTIPART_FLAGS \
+  (MU_MIME_MULTIPART_MIXED|MU_MIME_MULTIPART_ALT)
+
 int
 mu_mime_create (mu_mime_t *pmime, mu_message_t msg, int flags)
 {
-  mu_mime_t          mime = NULL;
-  int             ret = 0;
-  size_t          size;
-  mu_body_t          body;
-
+  mu_mime_t mime = NULL;
+  int ret = 0;
+  size_t size;
+  mu_body_t body;
+  
   if (pmime == NULL)
     return EINVAL;
+  
+  switch (flags & MIME_MULTIPART_FLAGS)
+    {
+    case MIME_MULTIPART_FLAGS:
+      return EINVAL;
+
+    case 0:
+      flags |= MU_MIME_MULTIPART_MIXED;
+    }
+  
   *pmime = NULL;
   if ((mime = calloc (1, sizeof (*mime))) == NULL)
     return ENOMEM;
@@ -886,10 +966,9 @@ mu_mime_create (mu_mime_t *pmime, mu_message_t msg, int flags)
     {
       if ((ret = mu_message_get_header (msg, &mime->hdrs)) == 0)
 	{
-	  if ((ret =
-	       mu_header_get_value (mime->hdrs,
-				    MU_HEADER_CONTENT_TYPE,
-				    NULL, 0, &size)) == 0 && size)
+	  if ((ret = mu_header_get_value (mime->hdrs,
+					  MU_HEADER_CONTENT_TYPE,
+					  NULL, 0, &size)) == 0 && size)
 	    {
 	      if ((mime->content_type = malloc (size + 1)) == NULL)
 		ret = ENOMEM;
@@ -922,7 +1001,7 @@ mu_mime_create (mu_mime_t *pmime, mu_message_t msg, int flags)
     }
   else
     {
-      mime->flags |= MIME_NEW_MESSAGE | MU_MIME_MULTIPART_MIXED;
+      mime->flags |= MIME_NEW_MESSAGE;
     }
   if (ret != 0)
     {
@@ -1027,9 +1106,10 @@ mu_mime_get_part (mu_mime_t mime, size_t part, mu_message_t *msg)
 	      if (ret == 0)
 		{
 		  mu_stream_set_flags (stream,
-				       MU_STREAM_READ | (flags &
-							 (MU_STREAM_SEEK
-							  | MU_STREAM_NONBLOCK)));
+				       MU_STREAM_READ
+				       | (flags &
+					  (MU_STREAM_SEEK
+					   | MU_STREAM_NONBLOCK)));
 		  mu_body_set_stream (body, stream, mime_part->msg);
 		  mu_message_set_body (mime_part->msg, body, mime_part);
 		  mime_part->body_created = 1;
@@ -1055,18 +1135,20 @@ mu_mime_get_num_parts (mu_mime_t mime, size_t *nmtp_parts)
 	    return (ret);
 	}
       else
-	mime->nmtp_parts = 1;
+	{
+	  *nmtp_parts = 1;
+	  return 0;
+	}
     }
   *nmtp_parts = mime->nmtp_parts;
   return (ret);
-
 }
 
 int
 mu_mime_add_part (mu_mime_t mime, mu_message_t msg)
 {
-  int             ret;
-
+  int ret;
+  
   if (mime == NULL || msg == NULL || (mime->flags & MIME_NEW_MESSAGE) == 0)
     return EINVAL;
   if ((ret = _mime_append_part (mime, msg, 0, 0, 0)) == 0)
@@ -1077,9 +1159,9 @@ mu_mime_add_part (mu_mime_t mime, mu_message_t msg)
 int
 mu_mime_get_message (mu_mime_t mime, mu_message_t *msg)
 {
-  mu_stream_t        body_stream;
-  mu_body_t          body;
-  int                ret = 0;
+  mu_stream_t body_stream;
+  mu_body_t body;
+  int ret = 0;
 
   if (mime == NULL || msg == NULL)
     return EINVAL;
@@ -1093,7 +1175,7 @@ mu_mime_get_message (mu_mime_t mime, mu_message_t *msg)
 	    {
 	      mu_message_set_header (mime->msg, mime->hdrs, mime);
 	      mu_header_set_value (mime->hdrs, MU_HEADER_MIME_VERSION, "1.0",
-				   0);
+				   1);
 	      if ((ret = _mime_set_content_type (mime)) == 0)
 		{
 		  if ((ret = mu_body_create (&body, mime->msg)) == 0)
@@ -1134,7 +1216,6 @@ mu_mime_to_message (mu_mime_t mime, mu_message_t *pmsg)
   if (rc == 0)
     {
       mu_message_unref (msg);
-      msg->flags |= MESSAGE_MIME_OWNER;
       mime->msg = NULL;
       mu_mime_ref (mime);
       *pmsg = msg;
