@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2006-2007, 2009-2012, 2014-2018 Free Software
-   Foundation, Inc.
+   Copyright (C) 2006-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
@@ -285,8 +284,13 @@ moderator_action (mu_sieve_machine_t mach)
       mu_sieve_abort (mach);
     }
 
-  mu_message_get_num_parts (msg, &nparts);
-
+  rc = mu_message_get_num_parts (msg, &nparts);
+  if (rc)
+    {
+      mu_sieve_error (mach, "mu_message_get_num_parts: %s", mu_strerror (rc));
+      mu_sieve_abort (mach);
+    }
+  
   if (nparts != 3) /* Mailman moderation requests have three parts */
     {
       mu_sieve_error (mach, _("expected 3 parts, but found %lu"),
