@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009-2012, 2014-2017 Free Software Foundation, Inc.
+   Copyright (C) 2009-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -52,248 +52,255 @@ static int set_folder (enum mailvar_cmd cmd,
 		       struct mailvar_variable *);
 static int set_headline (enum mailvar_cmd,
 			 struct mailvar_variable *);
-  
+
 struct mailvar_symbol mailvar_tab[] =
   {
     /* FIXME: */
-    { { "allnet", }, MAILVAR_HIDDEN },
-    
+    { { mailvar_name_allnet, }, MAILVAR_HIDDEN },
+
     /* For compatibility with other mailx implementations.
        Never used, always true. */
-    { { "append", },
+    { { mailvar_name_append, },
       MAILVAR_TYPEMASK (mailvar_type_boolean) | MAILVAR_RDONLY,
       N_("messages saved in mbox are appended to the end rather than prepended") },
-    { { "appenddeadletter", },
+    { { mailvar_name_appenddeadletter, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("append the contents of canceled letter to dead.letter file") },
-    { { "askbcc", },
+    { { mailvar_name_askbcc, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("prompt user for bcc before composing the message") },
-    { { "askcc", },
+    { { mailvar_name_askcc, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("prompt user for cc before composing the message") },
-    { { "ask", },
+    { { mailvar_name_ask, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("prompt user for subject before composing the message") },
-    { { "asksub", }, MAILVAR_ALIAS, NULL },
-    { { "autoinc", },
+    { { mailvar_name_asksub, }, MAILVAR_ALIAS, NULL },
+    { { mailvar_name_autoinc, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("automatically incorporate newly arrived messages")},
-    { { "autoprint", },
+    { { mailvar_name_autoprint, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("delete command behaves like dp") },
-    { { "byname" },
+    { { mailvar_name_byname },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("record outgoing messages in a file named after the first recipient; "
 	 "overrides the `record' variable") },
-    { { "bang", },
+    { { mailvar_name_bang, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("replace every occurrence of ! in arguments to the shell command"
 	 " with the last executed command") },
-    { { "charset", },
+    { { mailvar_name_charset, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("output character set for decoded header fields") },
-    { { "cmd", },
+    { { mailvar_name_cmd, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("default shell command for pipe") },
-    { { "columns", },
+    { { mailvar_name_columns, },
       MAILVAR_TYPEMASK (mailvar_type_number),
       N_("number of columns on terminal screen") },
-    { { "crt", },
+    { { mailvar_name_crt, },
       MAILVAR_TYPEMASK (mailvar_type_number) |
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("if numeric, sets the minimum number of output lines needed "
 	 "to engage paging; if boolean, use the height of the terminal "
 	 "screen to compute the threshold") },
-    { { "datefield", },
+    { { mailvar_name_datefield, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("get date from the `Date:' header, instead of the envelope") },
-    { { "debug", },
+    { { mailvar_name_debug, },
       MAILVAR_TYPEMASK (mailvar_type_string) |
-        MAILVAR_TYPEMASK (mailvar_type_boolean),
+	MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("set Mailutils debug level"),
       set_debug },
-    { { "decode-fallback", },
+    { { mailvar_name_decode_fallback, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("how to represent characters that cannot be rendered using the "
 	 "current character set"),
       set_decode_fallback },
-    { { "dot", },
+    { { mailvar_name_dot, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("input message is terminated with a dot alone on a line") },
-    { { "editheaders", },
+    { { mailvar_name_editheaders, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("allow editing message headers while composing") },
-    { { "emptystart", },
+    { { mailvar_name_emptystart, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("start interactive mode if the mailbox is empty") },
-    { { "escape", },
+    { { mailvar_name_escape, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("character denoting escapes") },
-    { { "flipr", },
+    { { mailvar_name_flipr, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("swap the meaning of reply and Reply commands") },
-    { { "folder", },
-      MAILVAR_TYPEMASK (mailvar_type_string), 
+    { { mailvar_name_folder, },
+      MAILVAR_TYPEMASK (mailvar_type_string),
       N_("folder directory name"),
       set_folder },
-    { { "fromfield", },
+    { { mailvar_name_fromfield, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("get sender address from the `From:' header, instead of "
 	 "the envelope") },
-    { { "gnu-last-command", },
+    { { mailvar_name_gnu_last_command, },
       MAILVAR_TYPEMASK (mailvar_type_string) | MAILVAR_RDONLY,
       N_("last executed command line") },
-    { { "header", },
+    { { mailvar_name_header, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("run the `headers' command after entering interactive mode") },
-    { { "headline", },
+    { { mailvar_name_headline, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("format string to use for the header summary"),
       set_headline },
-    { { "hold", },
+    { { mailvar_name_hold, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("hold the read or saved messages in the system mailbox") },
-    { { "ignore", },
+    { { mailvar_name_ignore, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("ignore keyboard interrupts when composing messages") },
-    { { "ignoreeof", },
+    { { mailvar_name_ignoreeof, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
-      N_("typing the EOF character terminates the letter being composed") },
-    { { "indentprefix", },
+      N_("ignore EOF character") },
+    { { mailvar_name_indentprefix, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("string used by the ~m escape for indenting quoted messages") },
-    { { "inplacealiases", },
+    { { mailvar_name_inplacealiases, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("expand aliases in the address header field "
-	 "before entering send mode") },
+	 "before starting composing the message") },
     /* For compatibility with other mailx implementations.
        Never used, always true. */
-    { { "keep", },
+    { { mailvar_name_keep, },
       MAILVAR_TYPEMASK (mailvar_type_boolean) | MAILVAR_RDONLY,
-      N_("truncate the user's system mailbox when it is empty") },
-    { { "keepsave", },
+      N_("keep the empty user's system mailbox,"
+	 " instead of removing it") },
+    { { mailvar_name_keepsave, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("keep saved messages in system mailbox too") },
-    { { "mailx", },
+    { { mailvar_name_mailx, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("enable mailx compatibility mode") },
-    { { "metamail", },
+    { { mailvar_name_metamail, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("interpret the content of message parts; if set to a string "
 	 "specifies the name of the external metamail command") },
-    { { "metoo", },
+    { { mailvar_name_metoo, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("do not remove sender addresses from the recipient list") },
-    { { "mimenoask", },
+    { { mailvar_name_mimenoask, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("a comma-separated list of MIME types for which "
 	 "no confirmation is needed before running metamail interpreter") },
-    { { "mode", },
+    { { mailvar_name_mode, },
       MAILVAR_TYPEMASK (mailvar_type_string) | MAILVAR_RDONLY,
       N_("the name of current operation mode") },
-    { { "nullbody", },
+    { { mailvar_name_nullbody, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("accept messages with an empty body") },
-    { { "nullbodymsg", },
+    { { mailvar_name_nullbodymsg, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("display this text when sending a message with empty body") },
-    { { "outfolder", },
+    { { mailvar_name_outfolder, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("keep created files in this folder") },
-    { { "page", },
+    { { mailvar_name_page, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
-      N_("pipe command terminates each message with a linefeed") },
-    { { "prompt", },
+      N_("pipe command terminates each message with a formfeed") },
+    { { mailvar_name_prompt, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("command prompt sequence") },
-    { { "quit", },
+    { { mailvar_name_quit, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("keyboard interrupts terminate the program") },
-    { { "rc", },
+    { { mailvar_name_rc, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("read the system-wide configuration file upon startup") },
-    { { "readonly", },
+    { { mailvar_name_readonly, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("mailboxes are opened in readonly mode") },
-    { { "record", },
+    { { mailvar_name_record, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("save outgoing messages in this file") },
-    { { "recursivealiases", },
+    { { mailvar_name_recursivealiases, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("recursively expand aliases") },
-    { { "regex", },
+    { { mailvar_name_regex, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
-      N_("use regular expressions in message specifications") },
-    { { "replyprefix", },
+      N_("use of regular expressions in search message specifications") },
+    { { mailvar_name_replyprefix, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("prefix for the subject line of a reply message") },
-    { { "replyregex", },
+    { { mailvar_name_replyregex, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("regexp for recognizing subject lines of reply messages"),
       set_replyregex },
-    { { "return-address" },
+    { { mailvar_name_return_address },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("return address for outgoing messages") },
-    { { "save", },
+    { { mailvar_name_save, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("store aborted messages in the user's dead.file") },
-    { { "screen", },
+    { { mailvar_name_screen, },
       MAILVAR_TYPEMASK (mailvar_type_number),
       N_("number of lines on terminal screen"),
       set_screen },
-    { { "sendmail", },
+    { { mailvar_name_sendmail, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("URL of the mail transport agent") },
     /* FIXME: Not yet used. */
-    { { "sendwait", },
+    { { mailvar_name_sendwait, },
       MAILVAR_TYPEMASK (mailvar_type_boolean) | MAILVAR_HIDDEN, NULL },
-    { { "sign", },
+    { { mailvar_name_sign, },
       MAILVAR_TYPEMASK (mailvar_type_string),
       N_("signature for use with the ~a command") },
-    { { "Sign", },
+    { { mailvar_name_Sign, },
       MAILVAR_TYPEMASK (mailvar_type_string),
-      N_("signature for use with the ~A command") },
-    { { "showenvelope", },
+      N_("name of the signature file for use with the ~A command") },
+    { { mailvar_name_showenvelope, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("`print' command includes the SMTP envelope in its output") },
-    { { "showto", },
+    { { mailvar_name_showto, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
-      N_("if the message was sent by the user, print its recipient address "
+      N_("if the message was sent by me, print its recipient address "
 	 "in the header summary") },
-    { { "toplines", },
+    { { mailvar_name_toplines, },
       MAILVAR_TYPEMASK (mailvar_type_number),
       N_("number of lines to be displayed by `top' or `Top'") },
-    { { "variable-pretty-print", },
+    { { mailvar_name_variable_pretty_print, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("print variables with short descriptions") },
-    { { "varpp", }, MAILVAR_ALIAS },
-    { { "variable-strict", },
+    { { mailvar_name_varpp, }, MAILVAR_ALIAS },
+    { { mailvar_name_variable_strict, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
-      N_("perform strict checking when setting options") },
-    { { "varstrict", }, MAILVAR_ALIAS },
-    { { "verbose", },
+      N_("perform strict checking when setting variables") },
+    { { mailvar_name_varstrict, }, MAILVAR_ALIAS },
+    { { mailvar_name_verbose, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("verbosely trace the process of message delivery"),
       set_verbose },
-    { { "xmailer", },
+    { { mailvar_name_xmailer, },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("add the `X-Mailer' header to the outgoing messages") },
-    
-    { { "mime" },
+
+    { { mailvar_name_mime },
       MAILVAR_TYPEMASK (mailvar_type_boolean),
       N_("always compose MIME messages") },
 
     /* These will be implemented later */
-    { { "onehop", }, MAILVAR_HIDDEN, NULL },
+    { { mailvar_name_onehop, }, MAILVAR_HIDDEN, NULL },
+
+    { { mailvar_name_quiet, },
+      MAILVAR_TYPEMASK (mailvar_type_boolean) | MAILVAR_HIDDEN,
+      N_("suppress the printing of the version when first invoked") },
+
+    { { mailvar_name_PID },
+      MAILVAR_TYPEMASK (mailvar_type_string) | MAILVAR_RDONLY,
+      N_("PID of this process") },
 
-    { { "quiet", }, MAILVAR_TYPEMASK (mailvar_type_boolean) | MAILVAR_HIDDEN,
-      "suppresses the printing of the version when first invoked" },
-    
     { { NULL }, }
   };
 
-static int mailvar_symbol_count = sizeof (mailvar_tab) / sizeof (mailvar_tab[0]) - 1;
+static int mailvar_symbol_count =
+  sizeof (mailvar_tab) / sizeof (mailvar_tab[0]) - 1;
 
 struct mailvar_symbol *
 find_mailvar_symbol (const char *var)
@@ -314,12 +321,11 @@ print_descr (mu_stream_t out, const char *s, int n,
 	     int doc_col, int rmargin, char *pfx)
 {
   mu_stream_stat_buffer stat;
-  
+
   if (!s)
     return;
-  
-  mu_stream_set_stat (out, MU_STREAM_STAT_MASK (MU_STREAM_STAT_OUT),
-		      stat);
+
+  mu_stream_set_stat (out, MU_STREAM_STAT_MASK (MU_STREAM_STAT_OUT), stat);
   stat[MU_STREAM_STAT_OUT] = n;
   do
     {
@@ -328,14 +334,14 @@ print_descr (mu_stream_t out, const char *s, int n,
 
       if (stat[MU_STREAM_STAT_OUT] && pfx)
 	mu_stream_printf (out, "%s", pfx);
-      
+
       while (stat[MU_STREAM_STAT_OUT] < doc_col)
 	mu_stream_write (out, " ", 1, NULL);
 
       for (p = s; *p && p < s + (rmargin - doc_col); p++)
 	if (mu_isspace (*p))
 	  space = p;
-      
+
       if (!space || p < s + (rmargin - doc_col))
 	{
 	  mu_stream_printf (out, "%s", s);
@@ -372,7 +378,7 @@ mailvar_find_variable (const char *name, int create)
 {
   struct mailvar_symbol *sym;
   struct mailvar_variable *var;
-  
+
   sym = find_mailvar_symbol (name);
   if (sym)
     var = &sym->var;
@@ -387,7 +393,7 @@ mailvar_find_variable (const char *name, int create)
 	  mu_list_create (&mailvar_list);
 	  mu_list_set_comparator (mailvar_list, mailvar_variable_comp);
 	}
-  
+
       if (mu_list_locate (mailvar_list, &entry, (void**)&p))
 	{
 	  if (!create)
@@ -419,11 +425,11 @@ mailvar_find_variable (const char *name, int create)
    If PTR is not NULL, it must point to
 
    int           if TYPE is mailvar_type_number or mailvar_type_boolean
-   const char *  if TYPE is mailvar_type_string. 
+   const char *  if TYPE is mailvar_type_string.
 
    Passing PTR=NULL may be used to check whether the variable is set
    without retrieving its value. */
-   
+
 int
 mailvar_get (void *ptr, const char *variable, enum mailvar_type type, int warn)
 {
@@ -453,7 +459,7 @@ mailvar_get (void *ptr, const char *variable, enum mailvar_type type, int warn)
       default:
 	break;
       }
-  
+
   return 0;
 }
 
@@ -470,14 +476,14 @@ mailvar_variable_reset (struct mailvar_variable *var)
 {
   if (!var->set)
     return;
-  
+
   switch (var->type)
     {
     case mailvar_type_string:
       free (var->value.string);
       var->value.string = NULL;
       break;
-	      
+
     default:
       break;
     }
@@ -493,7 +499,7 @@ mailvar_variable_reset (struct mailvar_variable *var)
 
    Unless MOPTF_QUIET bit is set in FLAGS, the function performs
    semantic check, using the builtin options table.
- 
+
    If VALUE is null the VARIABLE is unset. */
 int
 mailvar_set (const char *variable, void *value, enum mailvar_type type,
@@ -503,15 +509,17 @@ mailvar_set (const char *variable, void *value, enum mailvar_type type,
   const struct mailvar_symbol *sym = find_mailvar_symbol (variable);
   enum mailvar_cmd cmd =
     (flags & MOPTF_UNSET) ? mailvar_cmd_unset : mailvar_cmd_set;
-  
-  if (!(flags & MOPTF_QUIET) && mailvar_is_true ("variable-strict"))
+
+  if (!(flags & MOPTF_QUIET) && mailvar_is_true (mailvar_name_variable_strict))
     {
       if (!sym)
 	mu_diag_output (MU_DIAG_WARNING, _("setting unknown variable %s"),
 			variable);
       else if (sym->flags & MAILVAR_RDONLY)
 	{
-	  mu_error (_("Cannot set read-only variable %s"),
+	  mu_error (cmd == mailvar_cmd_set
+		     ? _("Cannot set read-only variable %s")
+		     : _("Cannot unset read-only variable %s"),
 		    variable);
 	  return 1;
 	}
@@ -532,7 +540,7 @@ mailvar_set (const char *variable, void *value, enum mailvar_type type,
   newvar.type = var->type;
   newvar.set = 0;
   memset (&newvar.value, 0, sizeof (newvar.value));
-  
+
   switch (cmd)
     {
     case mailvar_cmd_set:
@@ -543,7 +551,7 @@ mailvar_set (const char *variable, void *value, enum mailvar_type type,
 	    case mailvar_type_number:
 	      newvar.value.number = *(int*)value;
 	      break;
-	  
+
 	    case mailvar_type_string:
 	      {
 		char *p = strdup (value);
@@ -555,11 +563,11 @@ mailvar_set (const char *variable, void *value, enum mailvar_type type,
 		newvar.value.string = p;
 	      }
 	      break;
-	  
+
 	    case mailvar_type_boolean:
 	      newvar.value.bool = *(int*)value;
 	      break;
-		  
+
 	    default:
 	      abort();
 	    }
@@ -592,20 +600,33 @@ mailvar_set (const char *variable, void *value, enum mailvar_type type,
 static int
 set_folder (enum mailvar_cmd cmd, struct mailvar_variable *var)
 {
-  int rc = mu_set_folder_directory (var->value.string);
+  int rc;
+
+  if (var->value.string)
+    {
+      char *p = mu_tilde_expansion (var->value.string, MU_HIERARCHY_DELIMITER,
+				    NULL);
+      if (!p)
+	mu_alloc_die ();
+      if (var->set)
+	free (var->value.string);
+      var->value.string = p;
+    }
+
+  rc = mu_set_folder_directory (var->value.string);
   if (rc)
     mu_diag_funcall (MU_DIAG_ERROR, "mu_set_folder_directory",
 		     var->value.string, rc);
   return rc;
 }
-		   
+
 
 static int
 set_headline (enum mailvar_cmd cmd, struct mailvar_variable *var)
 {
   if (cmd == mailvar_cmd_unset)
     return 1;
-  
+
   mail_compile_headline (var->value.string);
   return 0;
 }
@@ -615,7 +636,7 @@ set_decode_fallback (enum mailvar_cmd cmd, struct mailvar_variable *var)
 {
   char *value;
   int rc;
-  
+
   switch (cmd)
     {
     case mailvar_cmd_set:
@@ -625,7 +646,7 @@ set_decode_fallback (enum mailvar_cmd cmd, struct mailvar_variable *var)
     case mailvar_cmd_unset:
       value = "none";
     }
-  
+
   rc = mu_set_default_fallback (value);
   if (rc)
     mu_error (_("Incorrect value for decode-fallback"));
@@ -738,7 +759,7 @@ mu_list_t
 mailvar_list_copy (int set)
 {
   mu_list_t list = NULL;
-  
+
   if (mailvar_list)
     mu_list_map (mailvar_list, mailvar_mapper, NULL, 1, &list);
   if (!list)
@@ -747,16 +768,25 @@ mailvar_list_copy (int set)
   mu_list_sort (list, mailvar_variable_comp);
   return list;
 }
-  
+
 
+enum
+  {
+    MAILVAR_ITR_ALL = 0,         /* Return all variables */
+    MAILVAR_ITR_SET = 0x1,       /* Return only variables that have been set */
+    MAILVAR_ITR_WRITABLE = 0x2   /* Return only writable variables */
+  };
+
 struct mailvar_iterator
 {
-  const char *prefix;
-  int prefixlen;
-  mu_list_t varlist;
-  mu_iterator_t varitr;
+  int flags;                /* MAILVAR_ITR_ bits */
+  const char *prefix;       /* Prefix to match */
+  int prefixlen;            /* Length of the prefix */
+  mu_list_t varlist;        /* List of collected variables */
+  mu_iterator_t varitr;     /* Iterator over varlist */
 };
-  
+
+/* Return next match from ITR */
 const char *
 mailvar_iterate_next (struct mailvar_iterator *itr)
 {
@@ -764,29 +794,60 @@ mailvar_iterate_next (struct mailvar_iterator *itr)
 
   while (!mu_iterator_is_done (itr->varitr))
     {
+      size_t len;
+
       mu_iterator_current (itr->varitr, (void**) &vp);
       mu_iterator_next (itr->varitr);
-      
+
+      if (itr->flags & MAILVAR_ITR_WRITABLE)
+	{
+	  const struct mailvar_symbol *sym = find_mailvar_symbol (vp->name);
+	  if (sym && (sym->flags & MAILVAR_RDONLY))
+	    continue;
+	}
+
       if (strlen (vp->name) >= itr->prefixlen
 	  && strncmp (vp->name, itr->prefix, itr->prefixlen) == 0)
-	return vp->name;
+	return strdup (vp->name);
+
+      /* See if it's a negated boolean */
+      if (itr->prefixlen >= 2 && memcmp (itr->prefix, "no", 2) == 0
+	  && (len = strlen (vp->name)) >= itr->prefixlen - 2
+	  && memcmp (vp->name, itr->prefix + 2, itr->prefixlen - 2) == 0)
+	{
+	  char *p;
+	  struct mailvar_symbol *sym = find_mailvar_symbol (vp->name);
+	  if (sym && !(sym->flags & (MAILVAR_TYPEMASK (mailvar_type_boolean))))
+	    continue;
+	  p = malloc (len + 3);
+	  if (p)
+	    {
+	      strcpy (p, "no");
+	      strcpy (p + 2, vp->name);
+	    }
+	  return p;
+	}
     }
   return NULL;
 }
 
+/* Initialize iterator, return the name of the first match */
 const char *
-mailvar_iterate_first (int set, const char *prefix, struct mailvar_iterator **pitr)
+mailvar_iterate_first (int flags,
+		       const char *prefix, struct mailvar_iterator **pitr)
 {
   struct mailvar_iterator *itr = mu_alloc (sizeof *itr);
+  itr->flags = flags;
   itr->prefix = prefix;
   itr->prefixlen = strlen (prefix);
-  itr->varlist = mailvar_list_copy (set);
+  itr->varlist = mailvar_list_copy (flags & MAILVAR_ITR_SET);
   mu_list_get_iterator (itr->varlist, &itr->varitr);
   mu_iterator_first (itr->varitr);
   *pitr = itr;
   return mailvar_iterate_next (itr);
 }
 
+/* Release memory used by ITR */
 void
 mailvar_iterate_end (struct mailvar_iterator **pitr)
 {
@@ -812,7 +873,7 @@ mailvar_printer (void *item, void *data)
 {
   struct mailvar_variable *vp = item;
   struct mailvar_print_closure *clos = data;
-  
+
   if (clos->prettyprint)
     {
       const struct mailvar_symbol *sym = find_mailvar_symbol (vp->name);
@@ -832,17 +893,17 @@ mailvar_printer (void *item, void *data)
     case mailvar_type_number:
       mu_stream_printf (clos->out, "%s=%d", vp->name, vp->value.number);
       break;
-	  
+
     case mailvar_type_string:
       mu_stream_printf (clos->out, "%s=\"%s\"", vp->name, vp->value.string);
       break;
-	  
+
     case mailvar_type_boolean:
       if (!vp->value.bool)
 	mu_stream_printf (clos->out, "no");
       mu_stream_printf (clos->out, "%s", vp->name);
       break;
-	  
+
     case mailvar_type_whatever:
       mu_stream_printf (clos->out, "%s %s", vp->name, _("oops?"));
     }
@@ -856,11 +917,11 @@ mailvar_print (int set)
   mu_list_t varlist;
   size_t count;
   struct mailvar_print_closure clos;
-  
+
   varlist = mailvar_list_copy (set);
   mu_list_count (varlist, &count);
   clos.out = open_pager (count);
-  clos.prettyprint = mailvar_is_true ("variable-pretty-print");
+  clos.prettyprint = mailvar_is_true (mailvar_name_variable_pretty_print);
   clos.width = util_screen_columns ();
 
   mu_list_foreach (varlist, mailvar_printer, &clos);
@@ -928,7 +989,7 @@ describe_symbol (mu_stream_t out, int width, const struct mailvar_symbol *sym)
     }
   mu_stream_printf (out, "\n");
   mu_stream_set_stat (out, 0, NULL);
-  
+
   mu_stream_printf (out, _("Type: "));
   for (i = 0, t = 0; i < sizeof (typestr) / sizeof (typestr[0]); i++)
     if (sym->flags & MAILVAR_TYPEMASK (i))
@@ -940,14 +1001,14 @@ describe_symbol (mu_stream_t out, int width, const struct mailvar_symbol *sym)
   if (!t)
     mu_stream_printf (out, "%s", gettext (typestr[0]));
   mu_stream_printf (out, "\n");
-  
+
   mu_stream_printf (out, "%s", _("Current value: "));
   mailvar_variable_format (out, &sym->var, _("[not set]"));
 
   if (sym->flags & MAILVAR_RDONLY)
     mu_stream_printf (out, " [%s]", _("read-only"));
   mu_stream_printf (out, "\n");
-  
+
   print_descr (out, gettext (sym->descr ? sym->descr : N_("Not documented")),
 	       1, 1, width - 1, NULL);
   mu_stream_printf (out, "\n");
@@ -959,7 +1020,7 @@ mail_variable (int argc, char **argv)
   int pagelines = util_get_crt ();
   int width = util_screen_columns ();
   mu_stream_t out = open_pager (pagelines + 1);
-  
+
   if (argc == 1)
     {
       struct mailvar_symbol *sym;
@@ -984,17 +1045,16 @@ mail_variable (int argc, char **argv)
   mu_stream_unref (out);
   return 0;
 }
-
 
 #ifdef WITH_READLINE
 static char *
-mailvar_generator (int set, const char *text, int state)
+mailvar_generator (int flags, const char *text, int state)
 {
   static struct mailvar_iterator *itr;
   const char *p;
-  
+
   if (!state)
-    p = mailvar_iterate_first (set, text, &itr);
+    p = mailvar_iterate_first (flags, text, &itr);
   else
     p = mailvar_iterate_next (itr);
 
@@ -1006,27 +1066,67 @@ mailvar_generator (int set, const char *text, int state)
   return strdup (p);
 }
 
+/* Completion generator for the "set" command */
 static char *
 mailvar_set_generator (const char *text, int state)
 {
-  return mailvar_generator (0, text, state);
+  return mailvar_generator (MAILVAR_ITR_WRITABLE, text, state);
 }
 
+/* Completion generator for the "unset" command */
 static char *
 mailvar_unset_generator (const char *text, int state)
 {
-  return mailvar_generator (1, text, state);
+  return mailvar_generator (MAILVAR_ITR_SET | MAILVAR_ITR_WRITABLE,
+			    text, state);
 }
 
-char **
-mailvar_set_compl (int argc, char **argv, int ws)
+/* Completion generator for the "variable" command */
+static char *
+mailvar_variable_generator (const char *text, int state)
 {
+  return mailvar_generator (MAILVAR_ITR_ALL, text, state);
+}
+
+/* Completion function for all three commands */
+char **
+mailvar_set_compl (int argc, char **argv, int point)
+{
+  /* Possible values for argv[0] are: set, unset, variable */
+  char **matches =
+    rl_completion_matches (point & COMPL_WS ? "" : argv[argc-1],
+			   argv[0][0] == 'u'
+			    ? mailvar_unset_generator
+			    : argv[0][0] == 's'
+			       ? mailvar_set_generator
+			       : mailvar_variable_generator);
   ml_attempted_completion_over ();
-  return rl_completion_matches (ws ? "" : argv[argc-1],
-				/* Possible values for argv[0] are:
-				   set, unset, variable */
-				argv[0][0] == 'u' ? mailvar_unset_generator
-				                  : mailvar_set_generator);
+  if (matches && matches[1] == NULL) /* Exact match */
+    {
+      switch (argv[0][0])
+	{
+	case 'u':
+	  ml_set_completion_append_character (' ');
+	  break;
+
+	case 's':
+	  {
+	    struct mailvar_symbol *sym = find_mailvar_symbol (matches[0]);
+	    if (sym &&
+		(sym->flags & (MAILVAR_TYPEMASK (mailvar_type_string)
+			       | MAILVAR_TYPEMASK (mailvar_type_number))))
+	      ml_set_completion_append_character ('=');
+	    else
+	      ml_set_completion_append_character (' ');
+	  }
+	  break;
+
+	default:
+	  ml_set_completion_append_character (0);
+	}
+    }
+
+  return matches;
 }
 
 #endif
