@@ -90,7 +90,6 @@ static msgset_t *msgset_select (int (*sel) (mu_message_t, void *),
 				     unsigned int max_matches);
 static int select_header (mu_message_t msg, void *closure);
 static int select_body (mu_message_t msg, void *closure);
-static int select_type (mu_message_t msg, void *closure);
 static int select_sender (mu_message_t msg, void *closure);
 static int select_deleted (mu_message_t msg, void *closure);
 static int check_set (msgset_t **pset);
@@ -103,8 +102,11 @@ static size_t message_count;
 static msgset_t *result;
 static mu_opool_t tokpool;
 
+typedef int (*message_selector_t) (mu_message_t, void *);
+static message_selector_t find_type_selector (int type);
+
 /* Line 371 of yacc.c  */
-#line 108 "msgset.c"
+#line 110 "msgset.c"
 
 # ifndef YY_NULL
 #  if defined __cplusplus && 201103L <= __cplusplus
@@ -159,7 +161,7 @@ extern int yydebug;
 typedef union YYSTYPE
 {
 /* Line 387 of yacc.c  */
-#line 57 "msgset.y"
+#line 59 "msgset.y"
 
   char *string;
   int number;
@@ -168,7 +170,7 @@ typedef union YYSTYPE
 
 
 /* Line 387 of yacc.c  */
-#line 172 "msgset.c"
+#line 174 "msgset.c"
 } YYSTYPE;
 # define YYSTYPE_IS_TRIVIAL 1
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
@@ -196,7 +198,7 @@ int yyparse ();
 /* Copy the second part of user declarations.  */
 
 /* Line 390 of yacc.c  */
-#line 200 "msgset.c"
+#line 202 "msgset.c"
 
 #ifdef short
 # undef short
@@ -492,12 +494,12 @@ static const yytype_int8 yyrhs[] =
 };
 
 /* YYRLINE[YYN] -- source line where rule number YYN was defined.  */
-static const yytype_uint8 yyrline[] =
+static const yytype_uint16 yyrline[] =
 {
-       0,    73,    73,    76,    80,    84,    88,    92,    96,   100,
-     106,   107,   111,   117,   123,   127,   133,   134,   140,   143,
-     158,   167,   181,   193,   196,   202,   203,   207,   213,   214,
-     228,   234,   237,   245,   255
+       0,    75,    75,    78,    82,    86,    90,    94,    98,   102,
+     108,   109,   113,   119,   125,   129,   135,   136,   142,   145,
+     160,   169,   184,   196,   199,   205,   206,   210,   216,   217,
+     231,   237,   240,   248,   258
 };
 #endif
 
@@ -1427,7 +1429,7 @@ yyreduce:
     {
         case 2:
 /* Line 1792 of yacc.c  */
-#line 73 "msgset.y"
+#line 75 "msgset.y"
     {
 	     result = msgset_make_1 (get_cursor ());
 	   }
@@ -1435,7 +1437,7 @@ yyreduce:
 
   case 3:
 /* Line 1792 of yacc.c  */
-#line 77 "msgset.y"
+#line 79 "msgset.y"
     {
 	     result = msgset_make_1 (get_cursor ());
 	   }
@@ -1443,7 +1445,7 @@ yyreduce:
 
   case 4:
 /* Line 1792 of yacc.c  */
-#line 81 "msgset.y"
+#line 83 "msgset.y"
     {
 	     result = (yyvsp[(1) - (1)].mset);
 	   }
@@ -1451,7 +1453,7 @@ yyreduce:
 
   case 5:
 /* Line 1792 of yacc.c  */
-#line 85 "msgset.y"
+#line 87 "msgset.y"
     {
 	     result = msgset_select (select_deleted, NULL, 0, 1);
 	   }
@@ -1459,7 +1461,7 @@ yyreduce:
 
   case 6:
 /* Line 1792 of yacc.c  */
-#line 89 "msgset.y"
+#line 91 "msgset.y"
     {
 	     result = msgset_select (select_deleted, NULL, 1, 1);
 	   }
@@ -1467,7 +1469,7 @@ yyreduce:
 
   case 7:
 /* Line 1792 of yacc.c  */
-#line 93 "msgset.y"
+#line 95 "msgset.y"
     {
 	     result = msgset_select (select_deleted, NULL, 0, total);
 	   }
@@ -1475,7 +1477,7 @@ yyreduce:
 
   case 8:
 /* Line 1792 of yacc.c  */
-#line 97 "msgset.y"
+#line 99 "msgset.y"
     {
 	     result = msgset_select (select_deleted, NULL, 1, 1);
 	   }
@@ -1483,7 +1485,7 @@ yyreduce:
 
   case 9:
 /* Line 1792 of yacc.c  */
-#line 101 "msgset.y"
+#line 103 "msgset.y"
     {
 	     result = msgset_select (select_deleted, NULL, 0, 1);
 	   }
@@ -1491,7 +1493,7 @@ yyreduce:
 
   case 11:
 /* Line 1792 of yacc.c  */
-#line 108 "msgset.y"
+#line 110 "msgset.y"
     {
 	     (yyval.mset) = msgset_append ((yyvsp[(1) - (3)].mset), (yyvsp[(3) - (3)].mset));
 	   }
@@ -1499,7 +1501,7 @@ yyreduce:
 
   case 12:
 /* Line 1792 of yacc.c  */
-#line 112 "msgset.y"
+#line 114 "msgset.y"
     {
 	     (yyval.mset) = msgset_append ((yyvsp[(1) - (2)].mset), (yyvsp[(2) - (2)].mset));
 	   }
@@ -1507,7 +1509,7 @@ yyreduce:
 
   case 13:
 /* Line 1792 of yacc.c  */
-#line 118 "msgset.y"
+#line 120 "msgset.y"
     {
 	     (yyval.mset) = (yyvsp[(1) - (1)].mset);
 	     if (check_set (&(yyval.mset)))
@@ -1517,7 +1519,7 @@ yyreduce:
 
   case 14:
 /* Line 1792 of yacc.c  */
-#line 124 "msgset.y"
+#line 126 "msgset.y"
     {
 	     (yyval.mset) = (yyvsp[(2) - (3)].mset);
 	   }
@@ -1525,7 +1527,7 @@ yyreduce:
 
   case 15:
 /* Line 1792 of yacc.c  */
-#line 128 "msgset.y"
+#line 130 "msgset.y"
     {
 	     (yyval.mset) = msgset_negate ((yyvsp[(2) - (2)].mset));
 	   }
@@ -1533,7 +1535,7 @@ yyreduce:
 
   case 17:
 /* Line 1792 of yacc.c  */
-#line 135 "msgset.y"
+#line 137 "msgset.y"
     {
 	     (yyval.mset) = msgset_expand ((yyvsp[(1) - (4)].mset), (yyvsp[(3) - (4)].mset));
 	     msgset_free ((yyvsp[(1) - (4)].mset));
@@ -1543,7 +1545,7 @@ yyreduce:
 
   case 19:
 /* Line 1792 of yacc.c  */
-#line 144 "msgset.y"
+#line 146 "msgset.y"
     {
 	     struct header_data hd;
 	     hd.header = (yyvsp[(1) - (2)].string);
@@ -1562,7 +1564,7 @@ yyreduce:
 
   case 20:
 /* Line 1792 of yacc.c  */
-#line 159 "msgset.y"
+#line 161 "msgset.y"
     {
 	     (yyval.mset) = msgset_select (select_body, (yyvsp[(1) - (1)].string), 0, 0);
 	     if (!(yyval.mset))
@@ -1575,14 +1577,15 @@ yyreduce:
 
   case 21:
 /* Line 1792 of yacc.c  */
-#line 168 "msgset.y"
+#line 170 "msgset.y"
     {
-	     if (strchr ("dnorTtu", (yyvsp[(1) - (1)].type)) == NULL)
+	     message_selector_t sel = find_type_selector ((yyvsp[(1) - (1)].type));
+	     if (!sel)
 	       {
 		 yyerror (_("unknown message type"));
 		 YYERROR;
 	       }
-	     (yyval.mset) = msgset_select (select_type, (void *)&(yyvsp[(1) - (1)].type), 0, 0);
+	     (yyval.mset) = msgset_select (sel, NULL, 0, 0);
 	     if (!(yyval.mset))
 	       {
 		 mu_error (_("No messages satisfy :%c"), (yyvsp[(1) - (1)].type));
@@ -1593,7 +1596,7 @@ yyreduce:
 
   case 22:
 /* Line 1792 of yacc.c  */
-#line 182 "msgset.y"
+#line 185 "msgset.y"
     {
 	     (yyval.mset) = msgset_select (select_sender, (void *)(yyvsp[(1) - (1)].string), 0, 0);
 	     if (!(yyval.mset))
@@ -1606,7 +1609,7 @@ yyreduce:
 
   case 23:
 /* Line 1792 of yacc.c  */
-#line 193 "msgset.y"
+#line 196 "msgset.y"
     {
 	     (yyval.string) = NULL;
 	   }
@@ -1614,7 +1617,7 @@ yyreduce:
 
   case 24:
 /* Line 1792 of yacc.c  */
-#line 197 "msgset.y"
+#line 200 "msgset.y"
     {
 	     (yyval.string) = (yyvsp[(1) - (1)].string);
 	   }
@@ -1622,7 +1625,7 @@ yyreduce:
 
   case 26:
 /* Line 1792 of yacc.c  */
-#line 204 "msgset.y"
+#line 207 "msgset.y"
     {
 	     (yyval.mset) = msgset_append ((yyvsp[(1) - (3)].mset), (yyvsp[(3) - (3)].mset));
 	   }
@@ -1630,7 +1633,7 @@ yyreduce:
 
   case 27:
 /* Line 1792 of yacc.c  */
-#line 208 "msgset.y"
+#line 211 "msgset.y"
     {
 	     (yyval.mset) = msgset_append ((yyvsp[(1) - (2)].mset), (yyvsp[(2) - (2)].mset));
 	   }
@@ -1638,7 +1641,7 @@ yyreduce:
 
   case 29:
 /* Line 1792 of yacc.c  */
-#line 215 "msgset.y"
+#line 218 "msgset.y"
     {
 	     if ((yyvsp[(3) - (3)].mset)->npart == 1)
 	       {
@@ -1656,7 +1659,7 @@ yyreduce:
 
   case 30:
 /* Line 1792 of yacc.c  */
-#line 229 "msgset.y"
+#line 232 "msgset.y"
     {
 	     (yyval.mset) = msgset_range ((yyvsp[(1) - (3)].number), total);
 	   }
@@ -1664,14 +1667,14 @@ yyreduce:
 
   case 31:
 /* Line 1792 of yacc.c  */
-#line 235 "msgset.y"
+#line 238 "msgset.y"
     {
 	   }
     break;
 
   case 32:
 /* Line 1792 of yacc.c  */
-#line 238 "msgset.y"
+#line 241 "msgset.y"
     {
 	     (yyval.mset) = msgset_expand ((yyvsp[(1) - (4)].mset), (yyvsp[(3) - (4)].mset));
 	     msgset_free ((yyvsp[(1) - (4)].mset));
@@ -1681,7 +1684,7 @@ yyreduce:
 
   case 33:
 /* Line 1792 of yacc.c  */
-#line 246 "msgset.y"
+#line 249 "msgset.y"
     {
 	     if ((yyvsp[(1) - (1)].number) > message_count)
 	       {
@@ -1695,7 +1698,7 @@ yyreduce:
 
   case 34:
 /* Line 1792 of yacc.c  */
-#line 256 "msgset.y"
+#line 259 "msgset.y"
     {
 	     (yyval.mset) = (yyvsp[(2) - (3)].mset);
 	   }
@@ -1703,7 +1706,7 @@ yyreduce:
 
 
 /* Line 1792 of yacc.c  */
-#line 1707 "msgset.c"
+#line 1710 "msgset.c"
       default: break;
     }
   /* User semantic actions sometimes alter yychar, and that requires
@@ -1935,7 +1938,7 @@ yyreturn:
 
 
 /* Line 2055 of yacc.c  */
-#line 260 "msgset.y"
+#line 263 "msgset.y"
 
 
 static int xargc;
@@ -2252,7 +2255,7 @@ msgset_expand (msgset_t *set, msgset_t *expand_by)
 }
 
 msgset_t *
-msgset_select (int (*sel) (mu_message_t, void *), void *closure, int rev,
+msgset_select (message_selector_t sel, void *closure, int rev,
 	       unsigned int max_matches)
 {
   size_t i, match_count = 0;
@@ -2406,35 +2409,147 @@ select_sender (mu_message_t msg, void *closure)
   free (sender);
   return status;
 }
-
-int
-select_type (mu_message_t msg, void *closure)
+
+static int
+select_type_d (mu_message_t msg, void *unused MU_ARG_UNUSED)
 {
-  int type = *(int*) closure;
-  mu_attribute_t attr= NULL;
+  mu_attribute_t attr;
 
-  mu_message_get_attribute (msg, &attr);
-
-  switch (type)
-    {
-    case 'd':
-      return mu_attribute_is_deleted (attr);
-    case 'n':
-      return mu_attribute_is_recent (attr);
-    case 'o':
-      return mu_attribute_is_seen (attr);
-    case 'r':
-      return mu_attribute_is_read (attr);
-    case 'u':
-      return !mu_attribute_is_read (attr);
-    case 't':
-      return mu_attribute_is_userflag (attr, MAIL_ATTRIBUTE_TAGGED);
-    case 'T':
-      return !mu_attribute_is_userflag (attr, MAIL_ATTRIBUTE_TAGGED);
-    }
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_deleted (attr);
   return 0;
 }
 
+static int
+select_type_n (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_recent (attr);
+  return 0;
+}
+
+static int
+select_type_o (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_seen (attr);
+  return 0;
+}
+
+static int
+select_type_r (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_read (attr);
+  return 0;
+}
+
+static int
+select_type_s (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_userflag (attr, MAIL_ATTRIBUTE_SAVED);
+  return 0;
+}
+
+static int
+select_type_t (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return mu_attribute_is_userflag (attr, MAIL_ATTRIBUTE_TAGGED);
+  return 0;
+}
+
+static int
+select_type_T (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return !mu_attribute_is_userflag (attr, MAIL_ATTRIBUTE_TAGGED);
+  return 0;
+}
+
+static int
+select_type_u (mu_message_t msg, void *unused MU_ARG_UNUSED)
+{
+  mu_attribute_t attr;
+
+  if (mu_message_get_attribute (msg, &attr) == 0)
+    return !mu_attribute_is_read (attr);
+  return 0;
+}
+
+struct type_selector
+{
+  int letter;
+  message_selector_t func;
+};
+
+static struct type_selector type_selector[] = {
+  { 'd', select_type_d },
+  { 'n', select_type_n },
+  { 'o', select_type_o },
+  { 'r', select_type_r },
+  { 's', select_type_s },
+  { 't', select_type_t },
+  { 'T', select_type_T },
+  { 'u', select_type_u },
+  { '/', NULL }, /* A pseudo-entry needed for msgtype_generator only */
+  { 0 }
+};
+
+static message_selector_t
+find_type_selector (int type)
+{
+  struct type_selector *p;
+  for (p = type_selector; p->func; p++)
+    {
+      if (p->letter == type)
+	return p->func;
+    }
+  return NULL;
+}
+
+#ifdef WITH_READLINE
+char *
+msgtype_generator (const char *text, int state)
+{
+  /* Allowed message types, plus '/'. The latter can folow a colon,
+     meaning body lookup */
+  static int i;
+  char c;
+
+  if (!state)
+    {
+      i = 0;
+    }
+  while ((c = type_selector[i].letter))
+    {
+      i++;
+      if (!text[1] || text[1] == c)
+	{
+	  char *s = mu_alloc (3);
+	  s[0] = ':';
+	  s[1] = c;
+	  s[2] = 0;
+	  return s;
+	}
+    }
+  return NULL;
+}
+#endif
+
 int
 select_deleted (mu_message_t msg, void *closure MU_ARG_UNUSED)
 {
