@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2005-2012, 2014-2017 Free Software Foundation,
-   Inc.
+   Copyright (C) 2003-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -127,15 +126,16 @@ copy_message (mu_mailbox_t mbox, size_t n, const char *file)
   return rc;
 }
 
+static struct mh_optinit optinit[] = {
+  { "draftfolder", "Draft-Folder" },
+  { "whatnowproc", "whatnowproc" },
+  { NULL }
+};
+
 int
 main (int argc, char **argv)
 {
-  mh_getopt (&argc, &argv, options, 0, args_doc, prog_doc, NULL);
-
-  if (!draftfolder)
-    draftfolder = mh_global_profile_get ("Draft-Folder", NULL);
-  if (!whatnowproc)
-    whatnowproc = mh_global_profile_get ("whatnowproc", NULL);
+  mh_getopt_ext (&argc, &argv, options, 0, optinit, args_doc, prog_doc, NULL);
 
   if (use_draft)
     draftmessage = "cur";
